/* /////////////////////////////////////////////////////////////////////////////
 * File:        comstl_olestring_functions.h (originally MOOleStr.h, ::SynesisCom)
 *
 * Purpose:     Contains classes and functions for dealing with OLE/COM strings.
 *
 * Created:     2nd March 1996
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/comstl
 *                          http://www.comstl.org/
 *
 *              email:      submissions@comstl.org  for submissions
 *                          admin@comstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file comstl_olestring_functions.h
///
/// Contains classes and functions for dealing with OLE/COM strings.

#ifndef COMSTL_INCL_H_COMSTL_OLESTRING_FUNCTIONS
#define COMSTL_INCL_H_COMSTL_OLESTRING_FUNCTIONS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define COMSTL_VER_H_COMSTL_OLESTRING_FUNCTIONS_MAJOR      2
# define COMSTL_VER_H_COMSTL_OLESTRING_FUNCTIONS_MINOR      5
# define COMSTL_VER_H_COMSTL_OLESTRING_FUNCTIONS_REVISION   3
# define COMSTL_VER_H_COMSTL_OLESTRING_FUNCTIONS_EDIT       123
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef COMSTL_INCL_H_COMSTL
# include "comstl.h"                     // Include the COMSTL root header
#endif /* !COMSTL_INCL_H_COMSTL */
#ifndef COMSTL_INCL_H_COMSTL_MEMORY_FUNCTIONS
# include "comstl_memory_functions.h"    //
#endif /* !COMSTL_INCL_H_COMSTL_MEMORY_FUNCTIONS */
#include <wchar.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The COMSTL components are contained within the comstl namespace. This is
 * actually an alias for stlsoft::comstl_project,
 *
 * The definition matrix is as follows:
 *
 * _STLSOFT_NO_NAMESPACE    _COMSTL_NO_NAMESPACE    comstl definition
 * ---------------------    --------------------    -----------------
 *  not defined              not defined             = stlsoft::comstl_project
 *  not defined              defined                 not defined
 *  defined                  not defined             comstl
 *  defined                  defined                 not defined
 *
 */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::comstl */
namespace comstl
{
# else
/* Define stlsoft::comstl_project */

namespace stlsoft
{

namespace comstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \defgroup comstl_olestring_functions OLE String Functions
/// \ingroup COMSTL
/// \brief These functions create, manipulate and destroy instances of COM's OLE string type
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Functions
 */

/// Creates an OLE string from an ANSI string
///
/// \param s The string from which to create the OLE string
/// \result The created OLE string
inline LPOLESTR olestring_create_a(cs_char_a_t const *s)
{
    LPOLESTR    posz;

    if(s == NULL)
    {
        posz = NULL;
    }
    else
    {
        cs_size_t   cch     =   ::MultiByteToWideChar(0, 0, s, -1, NULL, 0);

        posz    =   static_cast<LPOLESTR>(::CoTaskMemAlloc(sizeof(OLECHAR) * (cch + 1)));

        if(posz != NULL)
        {
            ::MultiByteToWideChar(0, 0, s, -1, posz, (cch + 1));
        }
    }

    return posz;
}

/// Creates an OLE string from a Unicode string
///
/// \param s The string from which to create the OLE string
/// \result The created OLE string
inline LPOLESTR olestring_create_w(cs_char_w_t const *s)
{
    LPOLESTR    posz;

    if(s == NULL)
    {
        posz = NULL;
    }
    else
    {
        cs_size_t   cch =   ::wcslen(s);

        posz    =   static_cast<LPOLESTR>(::CoTaskMemAlloc(sizeof(OLECHAR) * (cch + 1)));

        if(posz != NULL)
        {
            ::wcscpy(posz, s);
        }
    }

    return posz;
}

/// Creates an OLE string from an ANSI string
///
/// \param s The string from which to create the OLE string
/// \result The created OLE string
inline LPOLESTR olestring_create(cs_char_a_t const *s)
{
    return olestring_create_a(s);
}

/// Creates an OLE string from a Unicode string
///
/// \param s The string from which to create the OLE string
/// \result The created OLE string
inline LPOLESTR olestring_create(cs_char_w_t const *s)
{
    return olestring_create_w(s);
}

/// Destroys an OLE string
///
/// \param posz The OLE string to destroy
inline void olestring_destroy(LPOLESTR posz)
{
    comstl_message_assert("Invalid request to destroy non-COM string", CoTaskMemDidAlloc(posz) != 0);

    ::CoTaskMemFree(static_cast<LPVOID>(posz));
}

/// Duplicates an OLE string
///
/// \param posz The OLE string to duplicate
/// \return The copied OLE string
inline LPOLESTR olestring_dup(LPCOLESTR posz)
{
    return olestring_create(posz);
}

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_comstl_olestring_functions(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "COMSTL", "olestring_functions", __FILE__);

        LPOLESTR    str1   =   olestring_create("Hello, Sailor!");
        LPOLESTR    str2   =   olestring_create(L"Hello, Sailor!");

        if( NULL != str1 &&
            NULL != str2 &&
            0 != wcscmp(str1, str2))
        {
            r->report("OLE string (ANSI + Unicode) creation failed ", __LINE__);
            bSuccess = false;
        }

        LPOLESTR    str3   =   olestring_dup(str1);

        if( NULL != str1 &&
            NULL != str3 &&
            0 != wcscmp(str1, str3))
        {
            r->report("OLE string duplication failed ", __LINE__);
            bSuccess = false;
        }

        olestring_destroy(str3);
        olestring_destroy(str2);
        olestring_destroy(str1);

        return bSuccess;
    }

    unittest_registrar    unittest_comstl_olestring_functions(test_comstl_olestring_functions);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group comstl_olestring_functions

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace comstl
# else
} // namespace comstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !COMSTL_INCL_H_COMSTL_OLESTRING_FUNCTIONS */

/* ////////////////////////////////////////////////////////////////////////// */
