/* /////////////////////////////////////////////////////////////////////////////
 * File:        dotnetstl_string_accessor.h
 *
 * Purpose:     A useful tool for accessing a String object's content as a c-string.
 *
 * Created:     24th June 2003
 * Updated:     10th June 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/dotnetstl
 *                          http://www.dotnetstl.org/
 *
 *              email:      submissions@dotnetstl.org  for submissions
 *                          admin@dotnetstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file dotnetstl_string_accessor.h
///
/// A useful tool for accessing a String object's content as a c-string.

#ifndef DOTNETSTL_INCL_H_DOTNETSTL_STRING_ACCESSOR
#define DOTNETSTL_INCL_H_DOTNETSTL_STRING_ACCESSOR

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define DOTNETSTL_VER_H_DOTNETSTL_STRING_ACCESSOR_MAJOR    1
# define DOTNETSTL_VER_H_DOTNETSTL_STRING_ACCESSOR_MINOR    2
# define DOTNETSTL_VER_H_DOTNETSTL_STRING_ACCESSOR_REVISION 5
# define DOTNETSTL_VER_H_DOTNETSTL_STRING_ACCESSOR_EDIT     18
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef DOTNETSTL_INCL_H_DOTNETSTL
# include "dotnetstl.h"                     // Include the .netSTL root header
#endif /* !DOTNETSTL_INCL_H_DOTNETSTL */
#ifndef STLSOFT_INCL_H_STLSOFT_SAP_CAST
# include "stlsoft_sap_cast.h"
#endif /* !STLSOFT_INCL_H_STLSOFT_SAP_CAST */
#ifdef STLSOFT_UNITTEST
# include <wchar.h>
#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef _STLSOFT_NO_NAMESPACE
/* There is no stlsoft namespace, so must define ::dotnetstl */
namespace dotnetstl
{
#else
/* Define stlsoft::dotnet_project */

namespace stlsoft
{

namespace dotnetstl_project
{

#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// Template providing scoped C-string access to a String
template <ss_typename_param_k C>
class c_string_accessor
{
public:
    typedef C           char_type;
    typedef C           *pointer;
    typedef C const     *const_pointer;

// Construction
public:
    /// Construct from a System::String
    ///
    /// \param s The String for which C-string access is to be provided. Must not be NULL
    ss_explicit_k c_string_accessor(System::String *s)
        : m_s(get_cstring_(s))
    {}
    /// Release resources
    ~c_string_accessor()
    {
        System::Runtime::InteropServices::Marshal::FreeHGlobal(m_s);
    }

// Accessors
public:
    /// Implicit conversion operator to a C-string
    operator const_pointer() const
    {
        return m_s;
    }

// Implementation
private:
    pointer get_cstring_(System::String *s);

// Members
private:
    pointer m_s;

// Not to be implemented
private:
    c_string_accessor(c_string_accessor const &);
    c_string_accessor &operator =(c_string_accessor const &);
};

////////////////////////////////////////////////////////////////////////////////
// Implementation

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

template <>
inline c_string_accessor<ds_char_a_t>::pointer c_string_accessor<ds_char_a_t>::get_cstring_(System::String *s)
{
    return sap_cast<ds_char_a_t*>(System::Runtime::InteropServices::Marshal::StringToHGlobalAnsi(s).ToPointer());
}

template <>
inline c_string_accessor<ds_char_w_t>::pointer c_string_accessor<ds_char_w_t>::get_cstring_(System::String *s)
{
    return sap_cast<ds_char_w_t*>(System::Runtime::InteropServices::Marshal::StringToHGlobalUni(s).ToPointer());
}

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_dotnetstl_string_accessor(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, ".netSTL", "string_accessor", __FILE__);

        System::String  *s1 =   new System::String("String #1");
        System::String  *s2 =   new System::String("String #2");

        if(0 != strcmp(c_string_accessor<char>(s1), "String #1"))
        {
            r->report("c_string_accessor<char>() failed ", __LINE__);
            bSuccess = false;
        }

        if(0 != wcscmp(c_string_accessor<wchar_t>(s2), L"String #2"))
        {
            r->report("c_string_accessor<char>() failed ", __LINE__);
            bSuccess = false;
        }

        return bSuccess;
    }

    unittest_registrar    unittest_dotnetstl_string_accessor(test_dotnetstl_string_accessor);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef _STLSOFT_NO_NAMESPACE
} // namespace dotnetstl
#else
} // namespace dotnetstl_project
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* DOTNETSTL_INCL_H_DOTNETSTL_STRING_ACCESSOR */

/* ////////////////////////////////////////////////////////////////////////// */
