/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_auto_destructor.h (originally MLAutDtr.h, ::SynesisStd)
 *
 * Purpose:     Contains the auto_destructor and auto_array_destructor template
 *              classes.
 *
 * Created:     19th January 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_auto_destructor.h
///
/// Contains the auto_destructor and auto_array_destructor template classes.

#ifndef STLSOFT_INCL_H_STLSOFT_AUTO_DESTRUCTOR
#define STLSOFT_INCL_H_STLSOFT_AUTO_DESTRUCTOR

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_AUTO_DESTRUCTOR_MAJOR      1
# define STLSOFT_VER_H_STLSOFT_AUTO_DESTRUCTOR_MINOR      6
# define STLSOFT_VER_H_STLSOFT_AUTO_DESTRUCTOR_REVISION   2
# define STLSOFT_VER_H_STLSOFT_AUTO_DESTRUCTOR_EDIT       41
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

template <ss_typename_param_k T>
class auto_destructor;

template <ss_typename_param_k T>
class return_value_destructor;

template <ss_typename_param_k T>
class auto_array_destructor;

template <ss_typename_param_k T>
class return_value_array_destructor;

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class auto_destructor
/// This class acts as an automatic frame scope variable that manages
/// heap-allocated object instances.
///
/// \param T The value type
template <ss_typename_param_k T>
class auto_destructor
{
public:
    /// The value type
    typedef T                                               value_type;
    /// The current parameterisation of the type
    typedef auto_destructor<T>                              class_type;
    /// The return value type
    typedef return_value_destructor<T>                      return_value_type;
    /// The <a href = "http://synesis.com.au/resources/articles/cpp/movectors.pdf">move construction</a> type
    typedef stlsoft_define_move_rhs_type(return_value_type) return_value_rhs_type;

// Construction
public:
    /// Constructs from the pointer to an instance whose lifetime will be managedd
    ss_explicit_k auto_destructor(value_type *t)
        : m_value(t)
    {}
    /// Constructs from the return_value_type, transferring ownership from it
    auto_destructor(return_value_rhs_type rhs) // Note: not explicit
        : m_value(move_lhs_from_rhs(rhs).detach())
    {}
    /// Destroys the managed instance
    ~auto_destructor() stlsoft_throw_0()
    {
        delete m_value;
    }

// Operations
public:
    /// Detaches the managed instance from the auto_destructor and returns a pointer to it to the caller
    ///
    /// \note The caller becomes responsible for destroying the instance.
    value_type *detach()
    {
        value_type  *t = m_value;

        m_value = 0;

        return t;
    }

// Conversion
public:
    /// Implicit conversion to pointer to the managed instance
    operator value_type *()
    {
        return m_value;
    }
    /// Implicit conversion to pointer-to-const to the managed instance
    operator const value_type *() const
    {
        return m_value;
    }

// Members
private:
    value_type  *m_value;

// Not to be implemented
private:
    auto_destructor(class_type const &rhs);
    auto_destructor const &operator =(class_type const &rhs);
};

// class auto_array_destructor
/// This class acts as an automatic frame scope variable that manages
/// heap-allocated object arrays.
///
/// \param T The value type
template <ss_typename_param_k T>
class auto_array_destructor
{
public:
    /// The value type
    typedef T                                               value_type;
    /// The current parameterisation of the type
    typedef auto_array_destructor<T>                        class_type;
    /// The return value type
    typedef return_value_array_destructor<T>                return_value_type;
    /// The <a href = "http://synesis.com.au/resources/articles/cpp/movectors.pdf">move construction</a> type
    typedef stlsoft_define_move_rhs_type(return_value_type) return_value_rhs_type;

// Construction
public:
    /// Constructs from the pointer to the array whose lifetimes will be managedd
    ss_explicit_k auto_array_destructor(value_type t[])
        : m_value(t)
    {}
    /// Construct from an return_value_array_destructor<T>, transferring the managed instance from it
    auto_array_destructor(return_value_rhs_type rhs) // Note: not explicit
        : m_value(move_lhs_from_rhs(rhs).detach())
    {}
    /// Destroys the managed array
    ~auto_array_destructor() stlsoft_throw_0()
    {
        delete [] m_value;
    }

// Operations
public:
    /// Detaches the managed instance from the auto_array_destructor and returns a pointer to it to the caller
    ///
    /// \note The caller becomes responsible for destroying the instance.
    value_type *detach()
    {
        value_type  *t = m_value;

        m_value = 0;

        return t;
    }

// Conversion
public:
    /// Implicit conversion to pointer to the managed instance
    operator value_type *()
    {
        return m_value;
    }
    /// Implicit conversion to pointer-to-const to the managed instance
    operator const value_type *() const
    {
        return m_value;
    }

// Members
private:
    value_type  *m_value;

// Not to be implemented
private:
    auto_array_destructor(class_type const &rhs);
    auto_array_destructor const &operator =(class_type const &rhs);
};


// class return_value_destructor
/// This class acts as a return-value scope variable that manages
/// heap-allocated object instances.
///
/// \param T The value type
template <ss_typename_param_k T>
class return_value_destructor
{
public:
    /// The value type
    typedef T                                           value_type;
    /// The current parameterisation of the type
    typedef return_value_destructor<T>                  class_type;
    /// The auto type
    typedef auto_destructor<T>                          auto_type;
    /// The <a href = "http://synesis.com.au/resources/articles/cpp/movectors.pdf">move construction</a> type
    typedef stlsoft_define_move_rhs_type(class_type)    rhs_type;

// Construction
public:
#ifdef _STLSOFT_RETURN_VALUE_DESTRUCTOR_ENABLE_DIRECT_CTOR
    /// Constructs from the pointer to an instance whose lifetime will be managedd
    ss_explicit_k return_value_destructor(value_type *pt)
        : m_value(pt)
    {}
#endif /* _STLSOFT_RETURN_VALUE_DESTRUCTOR_ENABLE_DIRECT_CTOR */
    /// Construct from an auto_destructor<T>, transferring the managed instance from it
    return_value_destructor(auto_type &rhs) // Note: not explicit
        : m_value(rhs.detach())
    {}
    /// <a href = "http://synesis.com.au/resources/articles/cpp/movectors.pdf">Move constructor</a>
    return_value_destructor(rhs_type rhs) // Note: Move-constructor
        : m_value(move_lhs_from_rhs(rhs).detach())
    {}
    /// Destroys the managed instance
    ~return_value_destructor() stlsoft_throw_0()
    {
        // This fires a report if value is non-zero, which indicates that
        // the return value had not been used. This is arguably ok since this
        // only ever happens in debug builds (what would be the point in
        // including in a release build?), so its violation of the rules on
        // no-throw destructors can be forgiven.

#ifndef _STLSOFT_RETURN_VALUE_DESTRUCTOR_DISABLE_UNUSED_ASSERT
#if defined(__STLSOFT_COMPILER_IS_WATCOM)
        stlsoft_assert(m_value == 0);
#else
        stlsoft_message_assert("This return value was not used", m_value == 0);
#endif /* __STLSOFT_COMPILER_IS_WATCOM */
#endif /* !_STLSOFT_RETURN_VALUE_DESTRUCTOR_DISABLE_UNUSED_ASSERT */

        delete m_value;
    }

// Operations
private:
    friend class auto_destructor<T>;

    /// Detaches the managed instance from the return_value_destructor and returns a pointer to it to the caller
    ///
    /// \note The caller becomes responsible for destroying the instance.
    value_type *detach()
    {
        value_type  *t = m_value;

        m_value = 0;

        return t;
    }

// Members
private:
    value_type  *m_value;

// Not to be implemented
private:
    return_value_destructor const &operator = (class_type const &rhs);
};

// class return_value_array_destructor
/// This class acts as a return-value scope variable that manages
/// heap-allocated object arrays.
///
/// \param T The value type
template <ss_typename_param_k T>
class return_value_array_destructor
{
public:
    /// The value type
    typedef T                                           value_type;
    /// The current parameterisation of the type
    typedef return_value_array_destructor<T>            class_type;
    /// The auto type
    typedef auto_array_destructor<T>                    auto_type;
    /// The <a href = "http://synesis.com.au/resources/articles/cpp/movectors.pdf">move construction</a> type
    typedef stlsoft_define_move_rhs_type(class_type)    rhs_type;

// Construction
public:
#ifdef _STLSOFT_RETURN_VALUE_DESTRUCTOR_ENABLE_DIRECT_CTOR
    /// Constructs from the pointer to the array whose lifetimes will be managedd
    ss_explicit_k return_value_array_destructor(value_type t[])
        : m_value(pt)
    {}
#endif /* _STLSOFT_RETURN_VALUE_DESTRUCTOR_ENABLE_DIRECT_CTOR */
    /// Constructs from an auto_array_destructor<T> instance, transferring the managed array from it
    return_value_array_destructor(auto_type &rhs) // Note: not explicit
        : m_value(rhs.detach())
    {}
    /// <a href = "http://synesis.com.au/resources/articles/cpp/movectors.pdf">Move constructor</a>
    return_value_array_destructor(rhs_type rhs) // Note: Move-constructor
        : m_value(move_lhs_from_rhs(rhs).detach())
    {}
    /// Destroys the managed array
    ~return_value_array_destructor() stlsoft_throw_0()
    {
        // This fires a report if value is non-zero, which indicates that
        // the return value had not been used. This is arguably ok since this
        // only ever happens in debug builds (what would be the point in
        // including in a release build?), so its violation of the rules on
        // no-throw destructors can be forgiven.

#ifndef _STLSOFT_RETURN_VALUE_DESTRUCTOR_DISABLE_UNUSED_ASSERT
#if defined(__STLSOFT_COMPILER_IS_WATCOM)
        stlsoft_assert(m_value == 0);
#else
        stlsoft_message_assert("This return value was not used", m_value == 0);
#endif /* __STLSOFT_COMPILER_IS_WATCOM */
#endif /* !_STLSOFT_RETURN_VALUE_DESTRUCTOR_DISABLE_UNUSED_ASSERT */

        delete [] m_value;
    }

// Operations
private:
    friend class auto_array_destructor<T>;

    /// Detaches the managed instance from the return_value_array_destructor and returns a pointer to it to the caller
    ///
    /// \note The caller becomes responsible for destroying the instance.
    value_type *detach()
    {
        value_type  *t = m_value;

        m_value = 0;

        return t;
    }

// Members
private:
    value_type  *m_value;

// Not to be implemented
private:
    return_value_array_destructor const &operator = (class_type const &rhs);
};


/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_AUTO_DESTRUCTOR */

/* ////////////////////////////////////////////////////////////////////////// */
