/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_explicit_cast.h (Formerly MTExCast.h, ::SynesisStl)
 *
 * Purpose:     Class to provide explicit cast operators.
 *
 * Created:     20th September 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_explicit_cast.h
///
/// Class to provide explicit cast operators.

#ifndef STLSOFT_INCL_H_STLSOFT_EXPLICIT_CAST
#define STLSOFT_INCL_H_STLSOFT_EXPLICIT_CAST

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_EXPLICIT_CAST_MAJOR      1
# define STLSOFT_VER_H_STLSOFT_EXPLICIT_CAST_MINOR      2
# define STLSOFT_VER_H_STLSOFT_EXPLICIT_CAST_REVISION   2
# define STLSOFT_VER_H_STLSOFT_EXPLICIT_CAST_EDIT       15
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class explicit_cast

/// This class is used to provide explicit casting operators.
///
/// \param T The underlying type of the cast
template <ss_typename_param_k T>
class explicit_cast
{
public:
    typedef T                   value_type;
    typedef explicit_cast<T>    class_type;

// Construction
public:
    explicit_cast(T t)
        : m_t(t)
    {}

#ifndef __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT
    // For compilers that do not support partial specialisation we need to
    // enforce the constraint that only basic types may be used.
    ~explicit_cast()
    {
        union must_be_basic_type
        {
            int i;
            T   t;
        };
    }
#endif /* !__STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT */

// Conversions
public:
    operator value_type () const
    {
        return m_t;
    }

// Members
private:
    T   m_t;
};

#ifdef __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT
/// Being able to cast to non-const references does not work, since
/// the compilers will refuse to apply such to a temporary.
///
/// Furthermore, it is too evil to contemplate (see Effective C++ #30)
/// so we simply hide the constructor and conversion operator.
template <ss_typename_param_k T>
class explicit_cast<T &>
{
public:
    typedef T                   value_type;
    typedef explicit_cast<T>    class_type;

// Construction
private:
    explicit_cast(T &);

// Conversions
private:
    operator T & ();

# if defined(__STLSOFT_COMPILER_IS_GCC)
public: static void f() {}
# endif /* !__STLSOFT_COMPILER_IS_GCC */
};

# ifdef _STLSOFT_EXPLICIT_CAST_FORCE_ALLOW_REFERENCE_TO_CONST_UDT
/* Compiler check */
#  if defined(__STLSOFT_COMPILER_IS_BORLAND) || \
      defined(__STLSOFT_COMPILER_IS_DMC) || \
      defined(__STLSOFT_COMPILER_IS_GCC) || \
      (  defined(__STLSOFT_COMPILER_IS_MSVC) && \
        _MSC_VER == 1310) || \
      defined(__STLSOFT_COMPILER_IS_WATCOM)
  // These compilers are ok
#  elif defined(__STLSOFT_COMPILER_IS_INTEL) || \
      (  defined(__STLSOFT_COMPILER_IS_MSVC) && \
        _MSC_VER != 1310)
  // These are not
#   error The current compiler does not properly implement it
#  else
  // These are unknown
#   error The current compiler has not been assessed as to whether it correctly translates explicit_cast for references to non-const
#  endif /* compiler */

template <ss_typename_param_k T>
class explicit_cast<T const &>
{
public:
    typedef T                   value_type;
    typedef explicit_cast<T>    class_type;

// Construction
public:
    explicit_cast(T const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator T const & () const
    {
        return m_t;
    }

// Members
private:
    T const &m_t;
};
# else /* ? _STLSOFT_EXPLICIT_CAST_FORCE_ALLOW_REFERENCE_TO_CONST_UDT */

#  ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
#   if defined(__STLSOFT_COMPILER_IS_MSVC)
#    if _MSC_VER >= 1200
#     pragma warning(push)
#    endif /* _MSC_VER >= 1200 */
#    pragma warning(disable : 4512)
#   endif /* __STLSOFT_COMPILER_IS_MSVC */
#   include "stlsoft_explicit_cast_specialisations.h"
#   if defined(__STLSOFT_COMPILER_IS_MSVC)
#    if _MSC_VER >= 1200
#     pragma warning(pop)
#    else /* _MSC_VER */
#     pragma warning(disable : 4512)
#    endif /* _MSC_VER >= 1200 */
#   endif /* __STLSOFT_COMPILER_IS_MSVC */
#  endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

# endif /* _STLSOFT_EXPLICIT_CAST_FORCE_ALLOW_REFERENCE_TO_CONST_UDT */

/// Not sure I really like this one, and reserve the right to remove it
/// but for now it stays.
template <ss_typename_param_k T>
class explicit_cast<T *>
{
public:
    typedef T                   value_type;
    typedef explicit_cast<T>    class_type;

// Construction
public:
    explicit_cast(T *t)
        : m_t(t)
    {}

// Conversions
public:
    operator T * ()
    {
        return m_t;
    }

// Members
private:
    T   *m_t;
};

# ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

template <ss_typename_param_k T>
class explicit_cast<T const *>
{
public:
    typedef T                   value_type;
    typedef explicit_cast<T>    class_type;

// Construction
public:
    explicit_cast(T const *t)
        : m_t(t)
    {}

// Conversions
public:
    operator T const * () const
    {
        return m_t;
    }

// Members
private:
    T const *m_t;
};

# endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

#endif // __STLSOFT_CF_TEMPLATE_PARTIAL_SPECIALISATION_SUPPORT

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_EXPLICIT_CAST */

/* ////////////////////////////////////////////////////////////////////////// */
