/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_ptr_cast.h
 *
 * Purpose:     A cast that throws bad_cast for dynamic pointer casting, as well
 *              as references.
 *
 * Created:     28th December 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_ptr_cast.h
///
/// A cast that throws bad_cast for dynamic pointer casting, as well as references.

#ifndef STLSOFT_INCL_H_STLSOFT_PTR_CAST
#define STLSOFT_INCL_H_STLSOFT_PTR_CAST

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_PTR_CAST_MAJOR       1
# define STLSOFT_VER_H_STLSOFT_PTR_CAST_MINOR       2
# define STLSOFT_VER_H_STLSOFT_PTR_CAST_REVISION    2
# define STLSOFT_VER_H_STLSOFT_PTR_CAST_EDIT        8
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"               // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */
#ifndef STLSOFT_INCL_H_STLSOFT_TYPE_TRAITS
# include "stlsoft_type_traits.h"   // base_type_traits
#endif /* !STLSOFT_INCL_H_STLSOFT_TYPE_TRAITS */
#include <typeinfo>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// A cast that throws bad_cast for pointer cast failures, as well as for
/// reference casts.
template <typename Target>
struct ptr_cast
{
public:
    typedef Target                                              target_type;
    typedef ptr_cast<Target>                                    class_type;
    typedef typename stlsoft::base_type_traits<Target>::cv_type target_base_type;
    typedef target_base_type                                    &reference_type;
    typedef target_base_type                                    *pointer_type;

private:
    typedef typename stlsoft::base_type_traits<Target>::base_type   target_base_type_;
    static target_base_type_ *manage_const(target_base_type_ *p)
    {
        return p;
    }
#if defined(__STLSOFT_COMPILER_IS_BORLAND)
    static target_base_type_ *manage_const(target_base_type_ const *p)
    {
        return const_cast<target_base_type_ *>(p);
    }
#else
    static target_base_type_ const *manage_const(target_base_type_ const *p)
    {
        return p;
    }
#endif /* __STLSOFT_COMPILER_IS_BORLAND */

public:
    template <typename Source>
    ptr_cast(Source &s)
        : m_p(manage_const(&dynamic_cast<Target>(s)))
    {
        // Nothing to do: dynamic_cast throws for reference types
    }

    template <typename Source>
#if defined(__STLSOFT_COMPILER_IS_BORLAND) || \
    defined(__STLSOFT_COMPILER_IS_GCC) || \
    defined(__STLSOFT_COMPILER_IS_MWERKS)
    ptr_cast(Source *&s)
#else
    ptr_cast(Source *s)
#endif /* compiler */
        : m_p(manage_const(dynamic_cast<Target>(s)))
    {
        if(NULL == m_p)
        {
            throw std::bad_cast();
        }
    }

    ptr_cast(pointer_type pt)
        : m_p(pt)
    {}
    ptr_cast(reference_type t)
        : m_p(&t)
    {}

public:
    operator reference_type () const
    {
        return const_cast<reference_type>(*m_p);
    }
    operator pointer_type () const
    {
        return const_cast<pointer_type>(m_p);
    }

protected:
    pointer_type  m_p;

// Not to be implemented
private:
    ptr_cast(class_type const &);
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_PTR_CAST */

/* ////////////////////////////////////////////////////////////////////////// */
