/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_unittest.h
 *
 * Purpose:     Contains various simple self-contained algorithms.
 *
 * Created:     26th February 2004
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_unittest.h
///
/// Contains various simple self-contained algorithms.

#ifndef STLSOFT_INCL_H_STLSOFT_UNITTEST
#define STLSOFT_INCL_H_STLSOFT_UNITTEST

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_UNITTEST_MAJOR       1
# define STLSOFT_VER_H_STLSOFT_UNITTEST_MINOR       2
# define STLSOFT_VER_H_STLSOFT_UNITTEST_REVISION    2
# define STLSOFT_VER_H_STLSOFT_UNITTEST_EDIT        12
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"               // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

#ifndef STLSOFT_UNITTEST
# error This file should not be used when STLSOFT_UNITTEST is not defined. You should probably refrain from including it directly, since it is an internal implementation feature, and subject to change at any time
#endif /* !STLSOFT_UNITTEST */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{

namespace unittest
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Unit test interfaces
 */

/// Interface for the unit test reporter
struct unittest_reporter
{
public:
    virtual void set_project(char const *projectName) = 0;

    virtual void set_component(char const *componentName) = 0;

    virtual void set_file(char const *fileName) = 0;

    virtual void start() = 0;

    virtual void report(char const *message, int line = 0) = 0;

    virtual void stop() = 0;
};

typedef ss_bool_t (*unittest_function)(unittest_reporter *);

/// Interface for the unit test host
struct unittest_host
{
public:
    virtual ss_uint32_t register_unittest_fn(unittest_function ) = 0;

    virtual void deregister_unittest_fn(ss_uint32_t unittestId) = 0;
};

/* /////////////////////////////////////////////////////////////////////////////
 * Unit test functions
 */

/// \brief This function must be defined by the test program, and should return
/// an instance of a class implementing the unittest_host interface (protocol)
///
/// \note Because the STLSoft components use a scoping initialiser (Schwarz
/// counter) it is important that the implementing object is a heap instance, to
/// ensure that it is not destroyed prior to the destructors for the unit test
/// initialisers using it to deregister themselves.
extern "C" unittest_host *get_unittest_host(void);

/* /////////////////////////////////////////////////////////////////////////////
 * Unit test classes
 */

/// \brief This class is used to automatically register the unit-tests for a
/// given compilation unit with the unit test host.
///
/// To use it you simply declare an instance of it, inside a
class unittest_registrar
{
    typedef unittest_registrar  class_type;

public:
    unittest_registrar(unittest_function pfn)
        : m_host(get_unittest_host())
    {
        m_key = m_host->register_unittest_fn(pfn);
    }
    ~unittest_registrar()
    {
        if(0 != m_key)
        {
            m_host->deregister_unittest_fn(m_key);
        }
    }

private:
    unittest_host *const    m_host;
    ss_uint32_t             m_key;

// Not to be implemented
private:
    unittest_registrar(class_type const &rhs);
    unittest_registrar &operator =(class_type const &rhs);
};

/// \brief This class is used within unit tests to simplify the
/// process of registering the test information, and calling
/// \c start() and \c stop() on the reporter.
///
class unittest_initialiser
{
    typedef unittest_initialiser    class_type;

public:
    /// Constructor
    ///
    /// \param reporter The reporter to be used
    /// \param project The project name
    /// \param component The component name
    /// \param file The file name
    unittest_initialiser(   unittest_reporter   *reporter
                        ,   char const          *project
                        ,   char const          *component
                        ,   char const          *file)
        : m_reporter(reporter)
    {
        stlsoft_assert(NULL != reporter);
        stlsoft_assert(NULL != project);
        stlsoft_assert(NULL != component);
        stlsoft_assert(NULL != file);

        m_reporter->set_project(project);
        m_reporter->set_file(file);
        m_reporter->set_component(component);

        m_reporter->start();
    }
    ~unittest_initialiser()
    {
        m_reporter->stop();
    }

private:
    unittest_reporter *const    m_reporter;

// Not to be implemented
private:
    unittest_initialiser(class_type const &rhs);
    unittest_initialiser &operator =(class_type const &rhs);
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace unittest

using unittest::unittest_reporter;
using unittest::unittest_registrar;

} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */
