/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl_system_info.h
 *
 * Purpose:     Contains the basic_system_info class, which provides information
 *              regarding the host system, such as number of processors and
 *              page size.
 *
 * Created:     7th February 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_system_info.h
///
/// Contains the basic_system_info class, which provides information regarding the host system, such as number of processors and page size.

#ifndef WINSTL_INCL_H_WINSTL_SYSTEM_INFO
#define WINSTL_INCL_H_WINSTL_SYSTEM_INFO

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_SYSTEM_INFO_MAJOR      1
# define WINSTL_VER_H_WINSTL_SYSTEM_INFO_MINOR      8
# define WINSTL_VER_H_WINSTL_SYSTEM_INFO_REVISION   4
# define WINSTL_VER_H_WINSTL_SYSTEM_INFO_EDIT       31
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <winstl.h>             // Include the WinSTL root header

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \weakgroup libraries STLSoft Libraries
/// \brief The individual libraries

/// \weakgroup libraries_system System Library
/// \ingroup libraries
/// \brief This library provides facilities for accessing system attributes

/// \defgroup winstl_system_library System Library (WinSTL)
/// \ingroup WinSTL libraries_system
/// \brief This library provides facilities for accessing Win32 system attributes
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// Provides system information
///
/// This class wraps the GetSystemInfo() API function. Since the information that
/// this function provides is constant for any particular active system for its
/// lifetime, the function is called only once, as implemented via the
/// _get_systeminfo() method.
class system_info
{
public:
    typedef system_info class_type;

// Construction
private:
    system_info();
    ~system_info();

// Operations
public:
    /// Returns the number of processes on the host machine
    static ws_uint32_t number_of_processors()
    {
        return _get_systeminfo().dwNumberOfProcessors;
    }

    /// Returns the page size of the host machine
    static ws_size_t page_size()
    {
        return _get_systeminfo().dwPageSize;
    }

    /// Returns the allocation granularity of the host machine
    static ws_size_t allocation_granularity()
    {
        return _get_systeminfo().dwAllocationGranularity;
    }

    /// Provides a non-mutable (const) reference to the \c SYSTEM_INFO instance
    static const SYSTEM_INFO &get_systeminfo()
    {
        return _get_systeminfo();
    }

// Implementation
private:
    /// Unfortunately, something in this technique scares the Borland compilers (5.5
    /// and 5.51) into Internal compiler errors so the s_init variable in
    /// _get_systeminfo() is int rather than bool when compiling for borland.
    static SYSTEM_INFO &_get_systeminfo()
    {
        static SYSTEM_INFO  s_systeminfo;
#ifdef __STLSOFT_COMPILER_IS_BORLAND
        /* WSCB: Borland has an internal compiler error if use ws_bool_t */
        static ws_int_t     s_init = (::GetSystemInfo(&s_systeminfo), ws_true_v);
#else
        static ws_bool_t    s_init = (::GetSystemInfo(&s_systeminfo), ws_true_v);
#endif /* __STLSOFT_COMPILER_IS_BORLAND */

        STLSOFT_SUPPRESS_UNUSED(s_init);

        return s_systeminfo;
    }
};

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_winstl_system_info(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "WinSTL", "system_info", __FILE__);

#if 0
        if(<<TODO>>)
        {
            r->report("<<TODO>> failed ", __LINE__);
            bSuccess = false;
        }
#endif /* 0 */

        return bSuccess;
    }

    unittest_registrar    unittest_winstl_system_info(test_winstl_system_info);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WINSTL_INCL_H_WINSTL_SYSTEM_INFO */

/* ////////////////////////////////////////////////////////////////////////// */
