/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl_version_info.h (Originally in MWVerInf.h, ::SynesisWin)
 *
 * Purpose:     Helper for accessing version information.
 *
 * Created:     16th February 1998
 * Updated:     10th June 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 1999-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_version_info.h
///
/// Helper for accessing version information.

#ifndef WINSTL_INCL_H_WINSTL_VERSION_INFO
#define WINSTL_INCL_H_WINSTL_VERSION_INFO

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_VERSION_INFO_MAJOR     2
# define WINSTL_VER_H_WINSTL_VERSION_INFO_MINOR     1
# define WINSTL_VER_H_WINSTL_VERSION_INFO_REVISION  6
# define WINSTL_VER_H_WINSTL_VERSION_INFO_EDIT      54
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef WINSTL_INCL_H_WINSTL
# include "winstl.h"                        // Include the WinSTL root header
#endif /* WINSTL_INCL_H_WINSTL */
#ifndef WINSTL_INCL_H_WINSTL_PROCESSHEAP_ALLOCATOR
# include "winstl_processheap_allocator.h"  // winstl::processheap_allocator
#endif /* WINSTL_INCL_H_WINSTL_PROCESSHEAP_ALLOCATOR */
#ifdef __STLSOFT_CF_EXCEPTION_SUPPORT
# include <stdexcept>                       // std::exception
#endif /* __STLSOFT_CF_EXCEPTION_SUPPORT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \weakgroup libraries STLSoft Libraries
/// \brief The individual libraries

/// \weakgroup libraries_version Version Library
/// \ingroup libraries
/// \brief This library provides facilities for manipulating version APIs

/// \defgroup winstl_version_library Version Library (WinSTL)
/// \ingroup WinSTL libraries_version
/// \brief This library provides facilities for manipulating the Win32 version API
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

#ifdef __STLSOFT_CF_EXCEPTION_SUPPORT
/// Exception thrown by the version_info class
class version_info_exception
    : public winstl_ns_qual_std(exception)
{
};
#endif /* __STLSOFT_CF_EXCEPTION_SUPPORT */

/// Represents the fixed part of a version information block
class fixed_file_info
{
    typedef fixed_file_info class_type;

public:
    fixed_file_info(VS_FIXEDFILEINFO const *ffi)
        : m_ffi(ffi)
        , m_fileDateTime(calc_FileDateTime_(ffi))
    {}

public:
    ws_uint16_t ApiVerHigh() const
    {
        return HIWORD(m_ffi->dwStrucVersion);
    }
    ws_uint16_t ApiVerLow() const
    {
        return LOWORD(m_ffi->dwStrucVersion);
    }

    /// The major part of the file version 
    ws_uint16_t FileVerMajor() const
    {
        return HIWORD(m_ffi->dwFileVersionMS);
    }
    /// The minor part of the file version 
    ws_uint16_t FileVerMinor() const
    {
        return LOWORD(m_ffi->dwFileVersionMS);
    }
    /// The revision part of the file version 
    ws_uint16_t FileVerRevision() const
    {
        return HIWORD(m_ffi->dwFileVersionLS);
    }
    /// The build increment part of the file version 
    ws_uint16_t FileVerBuild() const
    {
        return LOWORD(m_ffi->dwFileVersionLS);
    }

    /// The major part of the product version 
    ws_uint16_t ProductVerMajor() const
    {
        return HIWORD(m_ffi->dwProductVersionMS);
    }
    /// The minor part of the product version 
    ws_uint16_t ProductVerMinor() const
    {
        return LOWORD(m_ffi->dwProductVersionMS);
    }
    /// The revision part of the product version 
    ws_uint16_t ProductVerRevision() const
    {
        return HIWORD(m_ffi->dwProductVersionLS);
    }
    /// The build increment part of the product version 
    ws_uint16_t ProductVerBuild() const
    {
        return LOWORD(m_ffi->dwProductVersionLS);
    }

    /// The file flags mask
    ws_uint32_t FileFlagsMask() const
    {
        return m_ffi->dwFileFlagsMask;
    }
    /// The file flags
    ws_uint32_t FileFlags() const
    {
        return m_ffi->dwFileFlags;
    }

    /// The file operating system
    ws_uint32_t FileOS() const
    {
        return m_ffi->dwFileOS;
    }

    /// The file type
    ws_uint32_t FileType() const
    {
        return m_ffi->dwFileType;
    }
    /// The file subtype
    ws_uint32_t FileSubtype() const
    {
        return m_ffi->dwFileSubtype;
    }

    /// The timestamp of the file
    FILETIME const &FileDateTime() const
    {
        return m_fileDateTime;
    }

private:
    static FILETIME calc_FileDateTime_(VS_FIXEDFILEINFO const *ffi)
    {
        FILETIME    ft = {  ffi->dwFileDateLS, ffi->dwFileDateMS };

        return ft;
    }

private:
    VS_FIXEDFILEINFO const *const   m_ffi;
    FILETIME const                  m_fileDateTime;

private:
    class_type &operator =(class_type const &);
};

/// Represents a variable file part of a version information block
class VsVarFileInfo
{
    typedef VsVarFileInfo   class_type;
public:
    /// Constructor
    VsVarFileInfo(void const *pv)
        : m_pv(pv)
    {}

private:
    void const *const   m_pv;

private:
    class_type &operator =(class_type const &);
};

/// Represents a variable string part of a version information block
class VsStringFileInfo
{
    typedef VsStringFileInfo    class_type;
public:
    /// Constructor
    VsStringFileInfo(void const *pv)
        : m_pv(pv)
    {}

private:
    void const *const m_pv;

private:
    class_type &operator =(class_type const &);
};


/// \brief Provides convenient access to aspects of a module's version information
///
class version_info
{
private:
    typedef processheap_allocator<ws_byte_t>    allocator_type;
public:
    typedef version_info                        class_type;

/// \name Construction
/// @{
public:
    /// Creates an instance corresponding to the version information from the given module
    ///
    /// \param moduleName The name of the module (.exe, .dll, etc.) to load
    ss_explicit_k version_info(ws_char_a_t const *moduleName)
        : m_pv(retrieve_module_info_block_(moduleName))
        , m_key(calc_key_(m_pv))
        , m_ffi(calc_ffi_(m_key))
        , m_children(calc_children_(m_ffi))
    {
#ifdef _DEBUG
        // Check that ffi is the same as the pointer returned from VerQueryValue("\\");
        VS_FIXEDFILEINFO    *ffi    =   NULL;
        UINT                cchInfo =   0;

        stlsoft_assert(::VerQueryValueA(m_pv, "\\", reinterpret_cast<void**>(&ffi), &cchInfo));
        stlsoft_assert(ffi == m_ffi);
#endif /* _DEBUG */


        fixed_file_info fixedInfo = FixedFileInfo();

#ifdef _DEBUG
        ws_uint16_t    j   =   fixedInfo.FileVerMajor();
        ws_uint16_t    n   =   fixedInfo.FileVerMinor();
        ws_uint16_t    r   =   fixedInfo.FileVerRevision();
        ws_uint16_t    b   =   fixedInfo.FileVerBuild();
#endif /* _DEBUG */
    }

    /// Creates an instance corresponding to the version information from the given module
    ///
    /// \param moduleName The name of the module (.exe, .dll, etc.) to load
    ss_explicit_k version_info(ws_char_w_t const *moduleName)
        : m_pv(retrieve_module_info_block_(moduleName))
        , m_key(calc_key_(m_pv))
        , m_ffi(calc_ffi_(m_key))
        , m_children(calc_children_(m_ffi))
    {
#ifdef _DEBUG
        // Check that ffi is the same as the pointer returned from VerQueryValue("\\");
        VS_FIXEDFILEINFO    *ffi    =   NULL;
        UINT                cchInfo =   0;

        stlsoft_assert(::VerQueryValueW(m_pv, L"\\", reinterpret_cast<void**>(&ffi), &cchInfo));
        stlsoft_assert(ffi == m_ffi);
#endif /* _DEBUG */


        fixed_file_info fixedInfo = FixedFileInfo();

#ifdef _DEBUG
        ws_uint16_t    j   =   fixedInfo.FileVerMajor();
        ws_uint16_t    n   =   fixedInfo.FileVerMinor();
        ws_uint16_t    r   =   fixedInfo.FileVerRevision();
        ws_uint16_t    b   =   fixedInfo.FileVerBuild();
#endif /* _DEBUG */
    }

    /// Releases any allocated resources
    ~version_info()
    {
        allocator_type().deallocate(static_cast<ws_byte_t*>(m_pv));
    }

/// @}

/// \name Properties
public:
    /// The length of the version information
    ws_size_t Length() const
    {
#ifndef __STLSOFT_CF_EXCEPTION_SUPPORT
        if(NULL == m_pv)
        {
            return 0;
        }
#endif /* !__STLSOFT_CF_EXCEPTION_SUPPORT */

        return *(static_cast<WORD*>(m_pv) + 0);
    }
    /// The length of the value part of the version block
    ws_size_t ValueLength() const
    {
#ifndef __STLSOFT_CF_EXCEPTION_SUPPORT
        if(NULL == m_pv)
        {
            return 0;
        }
#endif /* !__STLSOFT_CF_EXCEPTION_SUPPORT */

        return *(static_cast<WORD*>(m_pv) + 1);
    }
    /// The type field in the version block
    ws_size_t Type() const
    {
        winstl_assert(NULL != m_pv);

        return *(static_cast<WORD*>(m_pv) + 2);
    }
    /// The key of the version block
    wchar_t const *key() const
    {
        winstl_assert(NULL != m_pv);

        return m_key;
    }
    /// The FixedFileInfo part of the block
    fixed_file_info FixedFileInfo() const
    {
        winstl_assert(NULL != m_pv);

        return fixed_file_info(m_ffi);
    }
    /// The VarFileInfo part of the block
    VsVarFileInfo VarFileInfo() const
    {
        winstl_assert(NULL != m_pv);

        return VsVarFileInfo(NULL);
    }

private:
    static void *retrieve_module_info_block_(ws_char_a_t const *moduleName)
    {
        ws_dword_t  cb  =   ::GetFileVersionInfoSizeA(const_cast<ws_char_a_t*>(moduleName), NULL);
        void        *pv =   allocator_type().allocate(cb);

#ifndef __STLSOFT_CF_THROW_BAD_ALLOC
        if(pv == NULL)
        {
            return NULL;
        }
#endif /* !__STLSOFT_CF_THROW_BAD_ALLOC */

        if(!::GetFileVersionInfoA(const_cast<ws_char_a_t*>(moduleName), 0, cb, pv))
        {
            allocator_type().deallocate(static_cast<ws_byte_t*>(pv), cb);

#ifdef __STLSOFT_CF_EXCEPTION_SUPPORT
            throw version_info_exception();
#else /* ? __STLSOFT_CF_EXCEPTION_SUPPORT */
            pv = NULL;
#endif /* __STLSOFT_CF_EXCEPTION_SUPPORT */
        }

        return pv;
    }

    static void *retrieve_module_info_block_(ws_char_w_t const *moduleName)
    {
        ws_dword_t  cb  =   ::GetFileVersionInfoSizeW(const_cast<ws_char_w_t*>(moduleName), NULL);
        void        *pv =   allocator_type().allocate(cb);

#ifndef __STLSOFT_CF_THROW_BAD_ALLOC
        if(pv == NULL)
        {
            return NULL;
        }
#endif /* !__STLSOFT_CF_THROW_BAD_ALLOC */

        if(!::GetFileVersionInfoW(const_cast<ws_char_w_t*>(moduleName), 0, cb, pv))
        {
            allocator_type().deallocate(static_cast<ws_byte_t*>(pv), cb);

#ifdef __STLSOFT_CF_EXCEPTION_SUPPORT
            throw version_info_exception();
#else /* ? __STLSOFT_CF_EXCEPTION_SUPPORT */
            pv = NULL;
#endif /* __STLSOFT_CF_EXCEPTION_SUPPORT */
        }

        return pv;
    }

    static wchar_t const *calc_key_(void const *pv)
    {
#ifndef __STLSOFT_CF_EXCEPTION_SUPPORT
        if(NULL == pv)
        {
            return NULL;
        }
#endif /* !__STLSOFT_CF_EXCEPTION_SUPPORT */

        wchar_t const *key  =   reinterpret_cast<wchar_t const*>(static_cast<WORD const*>(pv) + 3);

        return key;
    }

    static VS_FIXEDFILEINFO const *calc_ffi_(wchar_t const *key)
    {
#ifndef __STLSOFT_CF_EXCEPTION_SUPPORT
        if(NULL == key)
        {
            return NULL;
        }
#endif /* !__STLSOFT_CF_EXCEPTION_SUPPORT */

        union
        {
            VS_FIXEDFILEINFO const  *ffi;
            ws_size_t               cb;
        } u;

        u.ffi = reinterpret_cast<VS_FIXEDFILEINFO const*>(&key[1 + wcslen(key)]);

        u.cb = (3 + u.cb) & ~3;

        return u.ffi;
    }

    static WORD const *calc_children_(VS_FIXEDFILEINFO const *ffi)
    {
#ifndef __STLSOFT_CF_EXCEPTION_SUPPORT
        if(NULL == ffi)
        {
            return NULL;
        }
#endif /* !__STLSOFT_CF_EXCEPTION_SUPPORT */

        union
        {
            WORD const  *pw;
            ws_size_t   cb;
        } u;

        u.pw    =   reinterpret_cast<WORD const*>(&ffi[1]);

        u.cb = (3 + u.cb) & ~3;

        return u.pw;
    }

private:
    void                    *const  m_pv;
    wchar_t const           *const  m_key;
    VS_FIXEDFILEINFO const  *const  m_ffi;
    WORD const              *const  m_children;

// Not to be implemented
private:
    version_info(class_type const &rhs);
    class_type &operator =(class_type const &rhs);
};

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group winstl_version_library

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WINSTL_INCL_H_WINSTL_VERSION_INFO */

/* ////////////////////////////////////////////////////////////////////////// */
